"""
Copyright 2021-2024 Vitaliy Zarubin

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
from enum import Enum
from pathlib import Path

import click.exceptions
from yaml import Loader
from yaml import load

from changeln.src.support.data.config import CHANGELOG_CONF
from changeln.src.support.data.template import CHANGELOG_TEMPLATE
from changeln.src.support.helper import get_path_file
from changeln.src.support.output import echo_stdout, echo_stderr
from changeln.src.support.texts import AppTexts

# Data versions
APP_NAME = 'changeln'
APP_VERSION = '2.0.0'

# Default path config
PATH_CONF = './.changeln/changeln.yaml'
PATH_MAKO = './.changeln/changeln.mako'


# Verbose output types
class OutType(Enum):
    markdown = 'markdown'
    html = 'html'
    pdf = 'pdf'

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self)


# Loader configuration yaml
class Conf:

    @staticmethod
    def get_app_name() -> str:
        return APP_NAME

    @staticmethod
    def get_app_version() -> str:
        return APP_VERSION

    @staticmethod
    def _get_path_conf(path, default):
        path = get_path_file(path)

        default = get_path_file(default, none=False)

        if path and str(path).lower().endswith('.yaml'):
            return path
        else:
            if not default.is_file():
                Conf._create_default_config(default)
            return Path(default)

    @staticmethod
    def _create_default_config(path):
        if not click.confirm(AppTexts.confirm_init()):
            exit(0)

        path_dir = os.path.dirname(path)

        # Create dir if not exist
        if not os.path.isdir(path_dir):
            Path(path_dir).mkdir()

        # Write default configuration file
        with open(path, 'w') as file:
            print(CHANGELOG_CONF, file=file)

        # Check and write default template file
        path_mako = get_path_file(PATH_MAKO, none=False)
        if not Path(path_mako).is_file():
            with open(path_mako, 'w') as file:
                print(CHANGELOG_TEMPLATE, file=file)

        echo_stdout(AppTexts.success_init(), 2)

    def __init__(self, path):
        # Get path config
        self.conf_path = Conf._get_path_conf(path, default=PATH_CONF)

        # Load config
        with open(self.conf_path, 'rb') as file:
            self.conf = load(file.read(), Loader=Loader)

    # Get config path
    def get_path_out(self) -> Path:
        path = Path(os.path.dirname(self.conf_path))
        if path.name == '.changeln':
            return Path(os.path.dirname(path))
        else:
            return path

    # Get template
    def get_template(self) -> str:
        if 'template' not in self.conf.keys():
            echo_stderr(AppTexts.error_load_key('template'))
            exit(1)
        else:
            if not isinstance(self.conf['template'], str):
                echo_stderr(AppTexts.error_load_key('template'))
                exit(1)
            path = get_path_file(self.conf['template'], none=False, starting=os.path.dirname(self.conf_path))
            if not path.is_file():
                echo_stderr(AppTexts.not_found_template())
                exit(1)
            return open(path, 'r').read()

    # Get parse key
    def get_parse(self) -> str:
        if 'parse' not in self.conf.keys():
            echo_stderr(AppTexts.error_load_key('parse'))
            exit(1)
        else:
            if not isinstance(self.conf['parse'], str):
                echo_stderr(AppTexts.error_load_key('parse'))
                exit(1)
            return self.conf['parse']

    # Get parse key
    def get_filter(self) -> str:
        if 'filter' not in self.conf.keys():
            echo_stderr(AppTexts.error_load_key('filter'))
            exit(1)
        else:
            if not isinstance(self.conf['filter'], str):
                echo_stderr(AppTexts.error_load_key('filter'))
                exit(1)
            return self.conf['filter']

    # Get parse key
    def get_commits(self) -> str:
        if 'commits' not in self.conf.keys():
            echo_stderr(AppTexts.error_load_key('commits'))
            exit(1)
        else:
            if not isinstance(self.conf['commits'], dict):
                echo_stderr(AppTexts.error_load_key('commits'))
                exit(1)
            return self.conf['commits']



import os
import re
from datetime import datetime

import git
from git import TagReference
from natsort import natsorted

from changeln.src.support.output import echo_stderr
from changeln.src.support.texts import AppTexts


class ParseGit:
    def __init__(self, commits: {}, parse_commit: str, filter_tags: str):
        self.commits = commits
        self.parse = parse_commit
        self.filter = filter_tags
        self.repo = git.Repo(os.getcwd())

    # Get list tags
    def get_tags(self) -> [TagReference]:
        tags: [TagReference] = []
        pattern = re.compile(self.filter)
        for tag in reversed(natsorted(self.repo.git.tag().split('\n'), key=lambda y: y.lower())):
            for item in self.repo.tags:
                if tag == item.name and (not self.filter or not pattern.match(tag)):
                    tags.append(item)
        return tags

    # Get latest tag
    def get_last_tag(self) -> str | None:
        tags = self.get_tags()
        if not tags:
            return None
        return str(tags[0])

    # Get count tags
    def get_count_tags(self) -> int:
        tags = self.get_tags()
        return len(tags)

    # Get commits by group
    def get_group_commits(self):
        # Add formats commit message
        def format_commit(commit_tag, commit) -> {}:
            search = None
            if self.parse:
                search = re.search(self.parse, commit.message)
            return dict(
                commit=commit,
                regex=search.groups() if search else [],
                clean=commit.message.split(commit_tag)[1].strip()
            )

        result = []
        for tag_commits in self._get_tag_commits():
            group_commits = []
            for key, value in self.commits.items():
                group_commits.append(dict(
                    name=key,
                    commits=[format_commit(value, commit) for commit in tag_commits['commits'] if
                             value in commit.message]
                ))
            result.append(dict(
                name=tag_commits['name'],
                date=tag_commits['date'],
                commits=tag_commits['commits'],
                group=group_commits,
            ))
        return result

    # Filter commits by tags from config commits
    def _filter_commit_name(self, commits: []):
        def check_tags(message: str) -> bool:
            for value in self.commits.values():
                if value in message:
                    return True
            return False

        return [commit for commit in commits if check_tags(commit.message)]

    # Split multiple tags in one commit
    def _split_multi_commits(self, commits: []):
        result = []
        pattern = '({})'.format(('|'.join(self.commits.values())
                                 .replace('[', r'\[')  # replace tag symbols
                                 .replace(']', r'\]')))
        for commit in commits:
            split = re.split(pattern, commit.message.strip())[1:]
            # One commit
            if len(split) == 2:
                result.append(commit)
            else:
                # Multiple commits
                index = 0
                while index < len(split) - 1:
                    message = '{} {}'.format(split[index].strip(), split[index + 1].strip())
                    result.append(git.Commit(
                        commit.repo,
                        commit.binsha,
                        tree=commit.tree,
                        author=commit.author,
                        authored_date=commit.authored_date,
                        author_tz_offset=commit.author_tz_offset,
                        committer=commit.committer,
                        committed_date=commit.committed_date,
                        committer_tz_offset=commit.committer_tz_offset,
                        message=message,
                        parents=commit.parents,
                        encoding=commit.encoding,
                        gpgsig=commit.gpgsig
                    ))
                    index += 2

        return result

    # Find commits in tags
    def _get_tag_commits(self):
        result = []
        tags = self.get_tags()
        if not tags:
            # Get list HEAD...BOTTOM
            try:
                result.append(dict(
                    name='HEAD',
                    date=datetime.now(),
                    commits=self._split_multi_commits(
                        commits=self._filter_commit_name(
                            commits=self.repo.iter_commits(no_merges=True)
                        )
                    )
                ))
            except ValueError:
                echo_stderr(AppTexts.error_empty_repo())
                exit(1)
        else:
            to = 'HEAD'
            # Get list HEAD...tag...tag
            for tag in tags:
                result.append(dict(
                    name=str(to),
                    date=datetime.now() if isinstance(to, str) else to.commit.committed_date,
                    commits=self._split_multi_commits(
                        commits=self._filter_commit_name(
                            commits=self.repo.iter_commits('{}...{}'.format(tag.name, to), no_merges=True)
                        )
                    )
                ))
                to = tag
            # Get list tag...BOTTOM
            result.append(dict(
                name=str(to),
                date=to.commit.committed_date,
                commits=self._split_multi_commits(
                    commits=self._filter_commit_name(
                        commits=self.repo.iter_commits(to, no_merges=True)
                    )
                )
            ))

        return result



"""
Copyright 2021-2024 Vitaliy Zarubin

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

CHANGELOG_CONF = r'''## Application configuration file Changeln
## Version config: 0.0.2

## Comment tags by which they will be searched
## and groups by which they will be analyzed.
commits:
  Bug: '[bug]'
  Change: '[change]'
  Feature: '[feature]'

## Path to mako template
template: ./changeln.mako

## Regular expression to break the comment into groups
## to format the comment output in the changelog in the 
## "regex" variable.
##
## The empty value is not used.
parse: ''

## Filter tags using regular expressions
##
## The empty value is not used.
filter: ''
'''



"""
Copyright 2021-2024 Vitaliy Zarubin

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import re
from datetime import datetime
from pathlib import Path

import markdown as mark
from mako.template import Template
from weasyprint import HTML

from changeln.src.support.conf import OutType
from changeln.src.support.output import echo_stdout
from changeln.src.support.parse_git import ParseGit
from changeln.src.support.texts import AppTexts


# Generate html from template mako
def _gen_changelog(ctx) -> str:
    """Common gen changelog."""
    parser = ParseGit(
        commits=ctx.obj.get_commits(),
        parse_commit=ctx.obj.get_parse(),
        filter_tags=ctx.obj.get_filter(),
    )

    ln_date = datetime.now()
    ln_last = parser.get_last_tag()
    ln_list_tags = parser.get_tags()
    ln_count_tags = parser.get_count_tags()
    ln_group_commits = parser.get_group_commits()

    out = Template(ctx.obj.get_template()).render(
        ln_date=ln_date,
        ln_last=ln_last,
        ln_list_tags=ln_list_tags,
        ln_count_tags=ln_count_tags,
        ln_group_commits=ln_group_commits,
    ).strip()

    return '\n'.join([line.strip() for line in re.sub(r'\n\n+', '\n\n', out).split('\n')])


def group_make(ctx: {}, output: str):
    """Generate changelog."""

    try:
        out_md = _gen_changelog(ctx)
    except Exception as e:
        echo_stdout(AppTexts.error_template_parse(e))
        exit(1)

    out_path = ctx.obj.get_path_out()

    match OutType(output):
        case OutType.markdown:
            out_path = out_path / 'CHANGELOG.md'
            with open(out_path, 'w') as file:
                print(out_md, file=file)
        case OutType.html:
            out_path = out_path / 'CHANGELOG.html'
            out_html = mark.markdown(out_md)
            with open(out_path, 'w') as file:
                print(out_html, file=file)
        case OutType.pdf:
            out_path = out_path / 'CHANGELOG.pdf'
            out_html = mark.markdown(out_md)
            HTML(string=out_html).write_pdf(out_path)

    echo_stdout(AppTexts.success_save_changelog(str(out_path)))



"""
Copyright 2021-2024 Vitaliy Zarubin

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import glob
import os
from pathlib import Path


# Find file in directory
def find_path_file(extension: str, path: Path) -> Path | None:
    files = glob.glob(f'{path}/*.{extension}')
    if files:
        return Path(files[0])
    return None


# Get full path file
def get_path_file(path: str, none: bool = True, starting: str = None) -> Path | None:
    if not path:
        return None

    if not starting:
        starting = os.getcwd()

    # Format path
    if path.startswith('~/'):
        path = os.path.expanduser(path)
    if path.startswith('./'):
        path = '{}{}'.format(starting, path[1:])
    if path.startswith('../'):
        path = '{}/{}'.format(starting, path)

    # Read path
    path = Path(path)

    if none and not path.is_file():
        return None

    return Path(path)



"""
Copyright 2021-2024 Vitaliy Zarubin

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from enum import Enum

import click
from bs4 import BeautifulSoup


# Color tags
class EchoColors(Enum):
    red = 'red'
    green = 'green'
    yellow = 'yellow'
    blue = 'blue'
    magenta = 'magenta'
    cyan = 'cyan'
    white = 'white'
    reset = 'reset'


# App output echo error
def echo_stderr(text: str, newlines: int = 1):
    if text:
        click.echo(_colorize_text(text).strip(), nl=False, err=True)
        for x in range(newlines):
            click.echo()


# App output echo
def echo_stdout(text: str, newlines: int = 1):
    if text:
        click.echo(_colorize_text(text).strip(), nl=False)
        for x in range(newlines):
            click.echo()


# App output echo just line
def echo_line(newlines: int = 1):
    for x in range(newlines):
        click.echo()


# Colorize text by tags
def _colorize_text(text: str) -> str:
    if '<' not in text:
        return text
    soup = BeautifulSoup(text, 'html.parser')
    for tag in EchoColors:
        for item in soup.findAll(tag.value):
            text = text.replace(
                '<{}>{}</{}>'.format(tag.value, item.text, tag.value),
                click.style(item.text, fg=tag.value))
    return click.style(text, fg='reset')



import os
import subprocess
from pathlib import Path

from changeln.src.support.output import echo_stderr
from changeln.src.support.texts import AppTexts


# Check dependency for init
def check_dependency_init():
    # Check git application
    try:
        subprocess.run(['git', '--version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except (Exception,):
        echo_stderr(AppTexts.error_dependency_git())
        exit(1)

    # Check folder has .git
    git_folder = Path(os.getcwd()) / '.git'
    if not git_folder.is_dir():
        echo_stderr(AppTexts.not_found_git_folder())
        exit(1)



"""
Copyright 2021-2024 Vitaliy Zarubin

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

CHANGELOG_TEMPLATE = r'''### Changeln template example
### Template engine is Mako - https://www.makotemplates.org/
### Base template format is Markdown - https://python-markdown.github.io/
###
### Show all values: 
### ${context.keys()}
### ${context.__dict__}
###
### Structure:
### ln_date: datetime
### ln_last: str
### ln_list_tags: [TagReference]
### ln_count_tags: int
### ln_group_commits: [
###     name: str, 
###     date: datetime,
###     commits: [Commit],
###     group: [
###             {
###                 name: str, 
###                 commits: [
###                     {
###                         commit: Commit, 
###                         regex: [], 
###                         clean: str
###                     }
###                 ]
###             }
###          ]
###     ]

<%! from datetime import datetime %>

${"##"} Updated: ${ln_date.strftime('%m/%d/%Y %H:%M:%S %p')}

${"##"} Info

- Last tag: ${ln_last}
- Released: ${ln_count_tags}

${"##"} Versions

% for item in ln_list_tags:
- Version: ${item.name} (${datetime.fromtimestamp(item.commit.committed_date).strftime('%d/%m/%Y')})
% endfor

% for tag in ln_group_commits:

    % if tag['commits']:
        % if tag['name'] == 'HEAD':
            ${"###"} HEAD (${ln_date.strftime('%d/%m/%Y')})
        % else:
            ${"###"} Version: ${tag['name']} (${datetime.fromtimestamp(tag['date']).strftime('%d/%m/%Y')})
        % endif
    % endif

    % for group in tag['group']:

        % if group['commits']:
            ${"####"} ${group['name']}

            % for commit in group['commits']:
                - ${commit['clean']} (${commit['commit'].author})
            % endfor
        % endif

    % endfor
% endfor
'''



"""
Copyright 2021-2024 Vitaliy Zarubin

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from enum import Enum


# Application texts
class AppTexts(Enum):
    @staticmethod
    def error_dependency_git():
        return '<red>Application</red> "git" <red>not found, install it.</red>'

    @staticmethod
    def error_template_parse(error: Exception):
        return '<red>Template parsing error.</red>\n{}: {}.'.format(type(error).__name__, error)

    @staticmethod
    def error_empty_repo():
        return '<red>Could not find comments, the repository may be empty.</red>'

    @staticmethod
    def confirm_init():
        return 'Add changeln configuration files to the project?'

    @staticmethod
    def success_init():
        return '<green>Configuration files added successfully.</green>'

    @staticmethod
    def success_save_changelog(path: str):
        return '<green>Changelog saved successfully:</green> {}'.format(path)

    @staticmethod
    def not_found_git_folder():
        return '<red>Changelog generation is not possible.</red> ".git" <red>directory not found.</red>'

    @staticmethod
    def not_found_template():
        return '<red>Template not found. Update the configuration file path template.</red>'

    @staticmethod
    def error_load_key(key: str):
        return '<red>Error reading configuration file. Check the</red> "{}" <red>parameter.</red>'.format(key)



import click

from changeln.src.features.group_make import group_make
from changeln.src.support.conf import Conf, OutType
from changeln.src.support.dependency import check_dependency_init

check_dependency_init()


@click.group(invoke_without_command=True)
@click.version_option(version=Conf.get_app_version(), prog_name=Conf.get_app_name())
@click.option(
    '--conf',
    default=None,
    help='Specify config path.',
    type=click.STRING,
    required=False)
@click.option(
    '--out',
    default=str(OutType.markdown),
    type=click.Choice([str(OutType.markdown), str(OutType.html), str(OutType.pdf)], case_sensitive=False),
    help='Type format output.')
@click.pass_context
def main(ctx: {}, conf: str, out: str):
    ctx.obj = Conf(conf)
    group_make(ctx, out)


if __name__ == '__main__':
    main(obj={})
